var path = require('path')
var fs = require('fs')
var exportExample = fs.readFileSync(require.resolve('../assets/resource/code-example-export.js')).toString()
	.replace(/^.*?(function[\s\S]+?\})\s*\(\s*function\([\s\S]+$/,"$1");
var exportRequireExample = fs.readFileSync(require.resolve('../assets/resource/code-example-export-require.js')).toString();
var ScriptLoader = require('./js-loader').ScriptLoader;
var Define = require('./define').Define;
var transformCode = require('./js-loader').transformCode

var maskCode = require('./exports-util').maskCode;
var parse64 = require('./exports-util').parse64;
var to64 = require('./exports-util').to64;
var makeFromCharCode = require('./exports-util').makeFromCharCode;
var makeFromToString = require('./exports-util').makeFromToString;

function formatOptions(options){
	var formats='r,f,c,raw,format,compressed';
	options = options||{};
	if(typeof options == 'string'){
		if(formats.split(',').indexOf(options)>=0){
			options = {format:options}
		}else{
			console.info('invalid format:'+options+'; supported format:'+formats)
		}
	}
	var f = options.format || options.f;
	if(f){
		if(options.format && formats.split(',').indexOf(options.format)<0){
			console.info('invalid format2:'+options.format+'; supported format:'+formats)
			//throw new Error();
		}
		options.format  = f;
	}
	var require = options.require;
	if(require){//export require
		options.require = !/^(0|false)$/i.test(require)
	}
	
	var mask = options.mask || options.m;
	
	if(mask){//export require
		options.mask = parseInt(mask);
	}
	//console.log(options)
	return options;
}
/**
 * @param onComplete(source,depFiles)
 * @param options{format:[raw,format,compressed]}
 */
function exportScript(base,paths,onComplete,options){
	if(arguments.length==3 && !(onComplete instanceof Function)){
		options = onComplete;
		onComplete = null;
	}
	options = formatOptions(options);
	//console.warn(2,options)
	//console.log('exportScript:',paths)
	//console.log(file,fs.existsSync(file) , fs.statSync(file).isFile())
	var file = require('path').join(base,paths[0]);
	if(paths.length==1 && fs.existsSync(file) && fs.statSync(file).isFile()){
		var isSingleJSFile =  paths.length == 1 && /\.js$/.test(paths[0]);
		var isSingleHTMLFile = paths.length == 1 && /\.html$/.test(paths[0]);
		//paths[0] = file.substr(base.length);
	}
	
	var base =fs.realpathSync(base||'./').replace(/\\/g,'/');
	var loader = new ScriptLoader(base,false);
	var loadFile = loader.loadFile
	var rtv ;
	var depMap = {};
	function onCompleteWrapper(value,error,fileDeps){
		rtv = value;
		fileDeps = [];
		for(var n in depMap){
			fileDeps.push(n);
		}
		//console.log('export deps files:',fileDeps)
		onComplete && onComplete.apply(this,arguments)
	}
	
	loader.loadFile = function(path,callback){
		loadFile.call(loader,path,function(text,path,file,deps){
			var i = deps && deps.length
			while(i--){
				depMap[deps[i]] =1;
			}
			callback.apply(this,arguments)
		})
	}
	
	
	
	resetPaths(loader,base,paths)
	//exportHTML(root,path,callback)
	if(isSingleHTMLFile){
		require('./export-html').exportHTML('./',paths[0], onCompleteWrapper)
	}else if(isSingleJSFile){
		//console.log('export single js file:'+paths)
		exportSingleFile(loader,paths[0], onCompleteWrapper,options)
	}else{
		//console.log(options)
		exportModules(loader,paths,onCompleteWrapper,options)
	}
	return rtv;
}
function exportModules(root,modules,callback,options){
	buildDefines(root,modules,function(defines,exportModules,internalDeps,fileDeps){//,externalDeps){
		/*
		 if(externalDeps.length >0){
			var error = 'has externalDeps not loaded!!, you must add all of the dependence modules in the exports list:'+externalDeps
			console.warn(error)
		}
		console.warn('externalDeps:',externalDeps)
		//*/
		//console.warn(exportModules,internalDeps)
		var exportOptions = {
			format:options.format,
			mask:options.mask,
			require:'require' in options?options.require:true,
			count:modules.length
		}
		//console.log(exportOptions)
		var result = buildExportSource(modules,defines,exportOptions);
		callback(result,null,fileDeps);
	});
}
/**
 * <ul>
 *   <li>1: 全依赖导出
 *     <p>只指定一个入口文件，该文件在全局域运行，变量全部是全局变量。&#32; 依赖模块可用 require 函数获得。</p></li>
 *   <li>2: 单库导出
 *     <p>不自动运行，需要require，内部用 require 函数获得外部模块, 该方式不自动合并外部模块的内容（只合并模块内文件）。</p></li>
 *   <li>3  匿名导出
 *     <p>同1，但全部子模块作为匿名闭包载入，不能再用require(module)方式获得</p></li>
 * </ul>
 */
function exportSingleFile(loader,file,callback,options){
	if(/#/.test(file)){//假文件
		var path = '#main'
		var content = file.substr(1);
		//var deps = Define.parseDependence(root,this.source);
		new Define(loader,path,function(path,realPath){
			var modules = this.deps;
			if(modules.length){
				function inlineCallback(defines,paths,internalDeps,externalDeps,fileDeps){
					paths.unshift('#main');
					defines.unshift(this)
					onExport(defines,paths,internalDeps,externalDeps,fileDeps)
				}
				buildDefines(loader,modules,inlineCallback);
			}else{
				content = transformCode(content,'#main',options.format);
				callback(content);
			}
		},content)
		
	}else{
		var modules = [file.replace(/^[\/\\]|\.js$/g,'')];
		//console.log(modules)
		buildDefines(loader,modules,onExport);
	}
	function onExport(defines,paths,internalDeps,externalDeps,fileDeps){
		if(internalDeps.indexOf([paths[0]])>=0){
			var error = 'export failed: main module is required by dependenced module!  try type=2';
			console.error(error)
			callback(null,error,fileDeps)
			return ;
		}
		var mainPath = paths.shift();
		var mainDefine = defines.shift();
		var replaceMap = {};
		for(var i=0;i<defines.length;i++){
			replaceMap[defines[i].path]=i;
		}
		var mainSource = mainDefine.toString(replaceMap);
		//console.log(mainSource)
		mainSource = mainSource.replace(/^[^{]+\{([\s\S]*)\}\s*\)\s*;?\s*/,'$1');
		if(/\bexports\s*\./.test(mainSource)){
			mainSource='var exports = exports || window;'+mainSource;
		}
		mainSource =transformCode (mainSource,mainPath,options.format);
		
		//console.warn(mainDefine.toString(),'\n\n####\n\n',mainSource)
		if(internalDeps.length + externalDeps.length ==0){
			callback(mainSource,null,fileDeps);
		}else{
			options.count = 0;//{count:0,ns:'',format:options.format}
			//'require' in options?options.require:true,
			if(!('require' in options)){
				options.require = true;
			}
			//console.error(options)
			var result = buildExportSource(paths,defines,options) + '\n'+mainSource;
			callback(result,null,fileDeps);
		}
	}
}

/**
 * @param root
 * @param exports(module_paths)
 * @param callback(defines,exportsPaths,internalDepPaths,externalDepPaths)
 */
function buildDefines(loader,modules,callback){
	var dependenceDepth = callback.length -2;//0:noneDeps,1:only directDeps,2:any deps;
	var addInternal = callback.length > 2;
	var addExternal = callback.length > 3;
	//console.log('buildDefines:',modules)
	var newExported = [];
	var exportsPaths=[],internalDepPaths=[],externalDepPaths = []
	var defines = [];
	var exportsMap = {};
	var count = modules.length;
	var loaded = 0;
	var prefixed = [];
	for(var i=0;i<count;i++){
		//console.error(modules[i])
		new Define(loader,modules[i],function(path,realPath){
			//console.log('@@1',path,realPath)
			exportsMap[path] = this;
			defines.push(this)
			exportsPaths.push(realPath);
			prefixed.push(realPath.replace(/\/.*/,''))
			
			if(++loaded == count){//为了保持顺序一致
				//console.log('prefixed:'+prefixed)
				if(addInternal){
					for(var j=0;j<count;j++){
						var def = defines[j]
						var depMap = def.depMap
						for(var n in depMap){
							var dep = depMap[n];
							if(!(dep in exportsMap)){
								if(newExported.indexOf(dep)==-1){
									var isInternal = prefixed.indexOf(dep.replace(/\/.*/,''))>=0;
									//console.log('isInternal:'+isInternal+':'+dep.replace(/\/.*/,''))
									if(addExternal || isInternal){
										//addExternal && console.log('add external:'+dep)
										newExported.push(dep);
									}
								}
							}
						}
					}
					loadNewExported();
				}else{
					callback(defines,exportsPaths);
				}
			}
			
		});
	}
	function loadNewExported(){
		var modules = newExported;
		var count = newExported.length;
		var loaded = 0;
		var baseIndex = defines.length
		if(count==0){
			callback(defines,exportsPaths,internalDepPaths,addExternal?externalDepPaths:null)
		}
		newExported = [];
		//console.log('@@####',modules)
		for(var i = 0;i<count;i++){
			new Define(loader,modules[i],function(path,realPath){
				if(path in exportsMap){
					console.error('double same require path:'+path)
				}
				exportsMap[path] = this;
				defines.push(this)
				var isInternal = prefixed.indexOf(realPath.replace(/\/.*/,''));
				(isInternal?internalDepPaths:externalDepPaths).push(realPath);
				if(++loaded == count){//为了保持顺序一致
					for(var j=0;j<count;j++){
						var def = defines[baseIndex+j]
						var depMap = def.depMap
						for(var n in depMap){
							var dep = depMap[n];
							//console.log('dep:',dep)
							if(!(dep in exportsMap)){
								if(newExported.indexOf(dep)==-1){
									var isInternal = prefixed.indexOf(dep.replace(/\/.*/,''))>=0;
									if(addExternal || isInternal){
										//addExternal && console.log('add external:'+dep)
										newExported.push(dep);
									}
								}
							}
						}
					}
					loadNewExported();;
				}
			});
		}
	}
}





function buildExportSource(paths,impls,exportConfig){
	exportConfig = exportConfig||{}
	
	console.log('buildExportSource:',exportConfig)
	var format = exportConfig.format;
	var exportCount = exportConfig.count || 0;
	var exportNS = exportConfig.ns || 'this';
	var exportRequire = !!exportConfig.require
	//
	var source = (exportRequire ? exportRequireExample:exportExample);
	//console.log(exportConfig,exportRequire,exportRequireExample.length , exportExample.length);
	//console.log(source)
	var replaceMap = {};
	for(var i=0;i<impls.length;i++){
		replaceMap[impls[i].path]=i;
	}
	//var exportNS = exportNS.replace(/^(this|window)$/,'');
	//console.log(source)
	if(exportCount){
		var dec = "";
		if(!/^(this|window)$/.test(exportNS)){
			if(/\./.test(exportNS)){
				dec = "var ns="+JSON.stringify(exportNS.split('.').reverse())+",current=this,i=ns.length;"+
				"while(i--){current = current[ns[i]] || (current[ns[i]] = {})}"
			}else{
				dec = "this."+exportNS+"=this.exportNS||{};";
			}
		}
		if(exportCount==1){
			var tail = "\tcopy(internal_require(0),"+exportNS+");\n"
			if(!source.match(/\bfunction\s+copy\(/)){
				tail+='\nfunction copy(src,o){for(i in src){o[i] = src[i];}}\n'
			}
			source = source.replace(/\}\s*$/,dec+tail+'}')
		}else if(exportCount>1){
			tail = "var i="+exportCount+";while(i--){copy(internal_require(i),"+exportNS+")}";
			if(!source.match(/\bfunction\s+copy\(/)){
				tail+='\nfunction copy(src,o){for(i in src){o[i] = src[i];}}\n'
			}
			source =  source.replace(/\}\s*$/,dec+tail+'}')
		}
	}
	
	impls = impls.map(function(def,i){
		//console.log(def.path)
		var code =  def.toString(replaceMap);
		
		return transformCode(code,def.path,format);
	})
	var mask = exportConfig.mask;
	
	if(mask){
		
		impls.map(function(code,i){
			//console.log(def.path)
			mask+=  code.length;
		});
		//32 ' ' 33:!,126:~,94
		//match function internal_require(){......impl(module.exports,...)
		var regexp = /\bfunction\s+(\w+)([\s\S]+?)\b(\w+(?:\[\w+\])?)(?=\((?:module\.exports\s*=|exports\s*,))/;
		
		//console.log(source,source.match(regexp))
		if(source.match(regexp)){
			/*
			code1.replace(/[\u0020-\u007e]/g,function(c){return mapTokens.charAt(c.charCodeAt()-32);}
			*/
			var mapTokens = String.fromCharCode.apply(String, Array(94+1+1).join(1).split('').map(
				function(a,i){return 126-i}
			).sort(function(a,b){return a&mask- b&mask}))
		}
	}
	
	
	var textCompress= 1;
	
	if(mapTokens){
		var hitMap = {};
		impls = impls.map(function(code){
			code = code.substring(code.indexOf('function('),code.lastIndexOf('}')+1);
			if(/\/\//.test(code)){
				//console.log(def.path,code)
			}
			//$JSI.define('lite/parse/template-token.js',[],function(exports,require){
			code = code.substring(code.indexOf('(')+1,code.lastIndexOf('}')).replace(/\/\/.*$/,'$&\n');
			return code;
		})
		impls.map(function(code){
			//console.log(code)
			//process.exit()
			code.replace(/([\w\$\_]+)|[^\w+\$]+/g,function(a,w){
				w+='$'
				if(w in hitMap){
					hitMap[w]++;
				}else{
					hitMap[w] = 1
				}
			})
		});
		
		var hits = [];
		for(var n in hitMap){
			hits.push([n.slice(0,-1),hitMap[n]])
		}
		hits.sort(function(a,b){return b[1] - a[1]})
		var tokenKeys = hits.map(function(t){return t[0]})
		tokenKeys = tokenKeys.slice(0,64).sort().concat(tokenKeys.slice(64).sort())
		tokenKeys = tokenKeys.sort();
		var char64=mapTokens.replace(/[^\w\$]/g,'');
		if(textCompress){
			impls = impls.map(function(code){
				return code.replace(/[$\w]+/g,function(w){
					try{
						return to64(tokenKeys.indexOf(w),char64)
					}catch(e){
						console.log(w,char64,tokenKeys.concat().sort().join(' '))
						throw e;
					}
				})
			});
			tokenKeys = tokenKeys.join(' ')
		}
		var source1Len = impls[0].length
		
			//.replace(/./g,function(c){return mapTokens.charAt(c.charCodeAt()-32);}),
			var prefix = '~function('
			var constructorPostfix = 'constructor';//'ructor'
			var constPostfixLen = constructorPostfix.length;
			var fromCharCodeLen = 'fromCharCode'.length
			var postfix = '}.apply(this,arguments)//fromCharCode'+constructorPostfix+'~function(';
			var postfixLen = postfix.length;
			
			var subsRadix = Math.random()>.5?35:36;
			var subsValue = parseInt('subs',subsRadix)
			
			var splitValue = parseInt('split',36)
			
			var replaceRadix = 28+ Math.round((36-28)*Math.random())
			var replaceValue = parseInt('replace',replaceRadix)
			
			var charRadix = 28+ Math.round((36-28)*Math.random())
			var charValue= parseInt('char',charRadix)
			var odeRadix = 25+ Math.round((36-25)*Math.random())
			var odeValue= parseInt('ode',odeRadix)
			//const,ructor
			//S,tring
			var vars = "fun,prefix,postfix='ing',charCodeAt_key='tr',mapTokens='CAS'," +
					"substr_key="+subsValue+",split_key="+splitValue+",replace_key="+replaceValue+"," +
					"exports_key,id_key," +
					"baseLength="+charValue+",toString_key='o',fromCharCode="+odeValue;
					
			var replaceTarget = ["\nvar ",vars.split(',').sort(function(a,b){return Math.random()-.5}).join(','),
			]
			if(textCompress){
				replaceTarget.push(
					"var token64,tokenKeys;",
					"function replace64(c){return tokenKeys[parse64(c,token64)]}",
					"function parse64(key,char64){var i=key.length;var v = 0;while(i){v<<=6;v+=char64.indexOf(key[--i]);}return v;}"
					)
			}
			replaceTarget.push("function replacer (c){return mapTokens[c[charCodeAt_key]()-32];}")
			replaceTarget.push("function $1$2"),
			replaceTarget.push("($3.length?(fun?fun:(" );
					
					replaceTarget.push("prefix =charCodeAt_key[0]+toString_key+mapTokens[2],")//toS
					replaceTarget.push("toString_key=prefix+charCodeAt_key+postfix,")//toString
					replaceTarget.push("substr_key = substr_key[toString_key]("+subsRadix+")+charCodeAt_key,");//substr
					replaceTarget.push("charCodeAt_key=baseLength[toString_key]("+charRadix+")+mapTokens[0]+fromCharCode[toString_key]("+odeRadix+")+mapTokens[1]+charCodeAt_key[0],");
					replaceTarget.push("split_key = split_key[toString_key](36),")
					replaceTarget.push("replace_key=replace_key[toString_key]("+replaceRadix+"),")
					
					//replaceTarget.push("x=console.log(toString_key,charCodeAt_key,substr_key,split_key,replace_key),")
					
					replaceTarget.push("mapTokens=$3[substr_key](-95)," )
			if(textCompress){
					replaceTarget.push('token64=mapTokens[replace_key](/[^\\w$]/g,""),')
			}
					replaceTarget.push("mapTokens=mapTokens[split_key](substr_key[substr_key](37))," )
					
					replaceTarget.push("postfix=$3[substr_key](-95-"+postfixLen+","+postfixLen+")[replace_key](/./g,replacer)," )
					replaceTarget.push("prefix=postfix[substr_key](-"+prefix.length+")," )
					
					replaceTarget.push("fun="+(constructorPostfix== 'constructor'?'':"'constructor'".slice(-constPostfixLen)+"'+")+"postfix[substr_key]("+ -(prefix.length+constPostfixLen)+","+constPostfixLen+"),")
					//replaceTarget.push("x=console.log(postfix,prefix,fun,$3[fun]),")
					replaceTarget.push("baseLength=$3.length,")
					replaceTarget.push("fromCharCode=postfix[substr_key]("+ -(prefix.length+constPostfixLen+fromCharCodeLen)+","+fromCharCodeLen+")," )
					//replaceTarget.push("x=console.error(fromCharCode),")
					replaceTarget.push("fromCharCode=fromCharCode[fun][fromCharCode],")
					replaceTarget.push("fun=$3[fun][fun]")
					//replaceTarget.push(",fun=function(s){console.log(s);return Function(s)}")
					replaceTarget.push(")):$3)")
					replaceTarget.push("(prefix+(" )
						//$3[replace_key](/[\u0020-\u007e]/g,replacer)
						if(textCompress){
							replaceTarget.push(	"tokenKeys?$3[replace_key](/[\u0020-\u007e]/g,replacer)[replace_key](/[\\w$]+/g,replace64):")
							replaceTarget.push(	"(" )
							replaceTarget.push(		"(tokenKeys=$3[replace_key](/[\u0020-\u007e]/g,replacer))" )
							replaceTarget.push(		"[substr_key]("+(tokenKeys.length+4)+","+source1Len+")" )
							replaceTarget.push(		"[replace_key](/[\\w$]+/g,(tokenKeys =tokenKeys[substr_key](2,"+tokenKeys.length+")[split_key](' '))&&replace64)" )
							replaceTarget.push(	")")
						}else{
							replaceTarget.push(	"$3[replace_key](/[\u0020-\u007e]/g,replacer)" )
						}
					replaceTarget.push(")+ '\\n'+postfix)");
			source = source.replace(regexp, replaceTarget.join('\n'));

//console.log(tokenKeys)
		
	}


	//console.log(source)
	
	
	if(mapTokens){
		//console.log(hits)
		impls = impls.map(function(code,i){
			var code1 = code+'//'+postfix;
			if(textCompress){
				code1 = '/*'+tokenKeys+'*/'+code1;
			}
			var code2 = i==0?maskCode(code1,mapTokens)+mapTokens:maskCode(code,mapTokens)
			
			//if(i==0){console.log(code,code2)}
			return JSON.stringify(code2);
		})
		var fromCharCodeInited = false;
		var baseLength = JSON.parse(impls[0]).length;
		//console.log(source)
		source=source.replace(/(fromCharCode=)|\.(exports|id|replace)\b|('\\n'|"\\n")/g,function(a,init,prop,str){
			
			//console.error('replace!!',a)
			if(!fromCharCodeInited){
				fromCharCodeInited = init;
			}else if(prop){
				if(/^[\da-z](2,8)$/.test(prop)  && Math.random()>0.5){
					var rtv = makeFromToString(prop,baseLength);
				}else{
					var rtv =  makeFromCharCode(prop,baseLength)
				}
				if(prop.match(/^(?:exports|replace|id)$/)){
					var rtv = prop+'_key || ('+prop+'_key='+rtv+')'
				}
				
				//console.error(rtv);
				return '['+rtv+']';
				
			}else if(str){
				var value = Function('return '+str)();
				return makeFromCharCode(value,baseLength)
			}
			return a;
		})
	}
	
	//console.log(format,source)
	source = transformCode(source.replace('__dirname','"./"')+'()','#export closure header',format).replace(/\(\);?$/,'');
	//console.log(mapTokens)
	if(exportRequire){//export require
		return source+"("+impls.join('\n,\n')+','+JSON.stringify(paths)+')'
	}else{
		if(mapTokens){
			//source = source.replace(/^\~function\(\)\{(.+)\}$/,function(a,s){return "~function(){eval("+JSON.stringify(s)+")}"})
			//console.log(source)
			
			return source + "("+impls.join(',')+')';
		}
		return source + "("+impls.join('\n,\n')+')';
	}
}


function resetPaths(loader,base,paths){
	outer:for(var i=0;i<paths.length;i++){
		var path = paths[i]
		if(path.charAt(0) == '#'){
			continue;
		}
		if(path.match(/^\w/)){
			path = './'+path
		}
		var absPath = require('path').join(fs.realpathSync(base),path).replace(/\\/g,'/');
		var bases = loader.bases;
		//console.warn(absPath)
		for(var j=0;j < bases.length;j++){
			var nbase = bases[j].base;
			//console.warn(nbase,absPath.indexOf(nbase))
			if(absPath.indexOf(nbase)==0){
				var path = absPath.substring(nbase.length).replace(/^node_modules\//,'');
				//console.error(path,nbase,loader.findModuleAndPath(path,nbase))
				
				paths[i] = loader.findModuleAndPath(path,nbase)[1];
				//console.log('ModuleAndPath',paths)
				continue outer;
			}
		}
	}
	//console.log(paths)
}
exports.buildDefines = buildDefines ;
exports.exportScript = exportScript;
